LoadPackage("atlasrep");
LoadPackage("io");

checkD := function(gr, c)
  local sg, x, y;
  x := Representative(c);
  for y in c do
    if not (x*y)^2 = (y*x)^2 then 
      sg := Subgroup(gr, [x,y]);
      if IsConjugate(sg, x, y) = false then
        return [x,y];
      fi;
    fi;
  od;
  return fail;
end;

x := [];
g := [];
n := 0;

name := "M12.2";

LogTo(Concatenation(name, ".log"));

M := Size(Maxes(CharacterTable(name)));

gr := AtlasGroup(name);
cc := ConjugacyClasses(gr);
ct := CharacterTable(name);

Print("Group: ", name, "\n");

ns := ClassPositionsOfNormalSubgroups(ct);
cl := Filtered([1..NrConjugacyClasses(ct)], x->not x in ns[2]);
Print("Conjugacy classes to study:\n");
for x in cl do
  Print(ClassNames(ct, "atlas")[x], ", ");
od;
Print("\n");
Display(List(cl, x->[ClassNames(ct, "atlas")[x], SizesConjugacyClasses(ct)[x], SizesCentralizers(ct)[x]]));
Print("==\n");

Print("Connecting group and character table: ", ConnectGroupAndCharacterTable(gr, ct), "\n");

bg := BaseOfGroup(gr);
Print("Base of Group: ", bg, "\n--\n");

f := IO_File(Concatenation(name, ".rec"), "w");
IO_WriteLine(f,"### The format for classinfo is [ r, s, order, comment ]");
IO_WriteLine(f,"data := rec(");
IO_WriteLine(f,"  base := ", bg, ",");
IO_WriteLine(f,"  group := \"", name, "\",");
IO_WriteLine(f,"  comments := \"Maximal subgroups\",");
IO_WriteLine(f,"  classinfo := [");


rem := [];

# For  an  ordinary character table <tbl> with known underlying group $G$,
# IdentificationOfConjugacyClasses returns a list of positive integers that
# contains at position i the position of the i-th conjugacy class of <tbl>
# in the list ConjugacyClasses(G).
for n in cl do
  c := cc[IdentificationOfConjugacyClasses(ct)[n]];
  g := Representative(c);
  Print("Checking: ", ClassNames(ct, "atlas")[n], "\n");
  Print("Size: ", SizesConjugacyClasses(ct)[n], "\n");
  Print("Size of the centralizer: ", SizesCentralizers(ct)[n], "\n");
  rs := checkD(gr, c); 
  if not rs = fail then
  	Print("r:=", OnTuples(bg, rs[1]), ";\n");
  	Print("s:=", OnTuples(bg, rs[2]), ";\n");
    IO_WriteLine(f, "    [ ", OnTuples(bg, rs[1]), ", ", OnTuples(bg, rs[2]), ", ", Order(g), ", \"D\" ],");
    Print("Removed: ", ClassNames(ct, "atlas")[n], "\n");
    Add(rem, n);
    Print("==\n");
  else
    Print("Not of type D!\n");
    Print("==\n");
  fi;
od;

a := List(cl, x->ClassNames(ct,"atlas")[x]);
b := List(rem, x->ClassNames(ct,"atlas")[x]);

Print("Summary:\n");
Print("To study: ", a, "\n");
Print("Removed: ", b, "\n");
Print("Not of type D: ", Difference(a,b), "\n");

IO_WriteLine(f, "  ]");
IO_WriteLine(f, ");");
IO_Flush(f);
IO_Close(f);

